# -*- coding: utf-8 -*-
##############################################################################
#
#    school module for OpenERP
#    Copyright (C) 2010 Tecnoba S.L. (http://www.tecnoba.com)
#       Pere Ramon Erro Mas <pereerro@tecnoba.com> All Rights Reserved.
#
#    This file is a part of school module
#
#    school OpenERP module is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    school OpenERP module is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv, fields

_MIN_DATETIME='1970-01-01 00:00:00'
_MAX_DATETIME='2099-12-31 23:59:59'

def agafa_id(unid):
    if type(unid)==type(tuple()):
        return unid[0]
    return unid

class groups_group(osv.osv):
    _name = 'groups.group'

groups_group()

class groups_participation(osv.osv):
    _name = 'groups.participation'

groups_participation()

class groups_group_assignation(osv.osv):
    _name = 'groups.group_assignation'
    _description = 'A assignation group in a period'

    _columns =  {
        'group_id' : fields.many2one('groups.group','Group', required=True,
            select=1, ondelete='cascade',),
        'participation_id' : fields.many2one('groups.participation', 'Participation',
            required=True, select=1, ondelete='cascade'),
        'datetime_from' : fields.datetime('Begining', select=1,),
        'datetime_to' : fields.datetime('End', select=1,)
    }

    _sql_constraints = [('dates_ok','CHECK (datetime_to>=datetime_from)','Date to minor than date from'),]


    def _assignation_in_domain(self, cr, uid, ids):
        data=self.read(cr, uid, ids, ['group_id','participation_id','datetime_from','datetime_to'], context={'withoutBlank': False})
        gids=set()
        for item in data:
            if item['datetime_from']!=item['datetime_to']: gids.add(item['group_id'])
        domains=self.pool.get('groups.group')._domain_assignations(cr, uid, list(gids))
        if domains:
            for item in data:
                group_id=item['group_id']
                participation_id=item['participation_id']
                if (group_id not in domains) or (domains[group_id] is None): continue
                # No participation in domain group
                if not participation_id in domains[group_id]: return False
                interval_into=(item['datetime_from'],item['datetime_to'])
                ret=False
                for interval_domain in domains[group_id][participation_id]:
                    if interval_into[0]>=interval_domain[0] and interval_into[1]<=interval_domain[1]:
                        ret=True
                # No inclusive intervals in domain group for participation_id
                if not ret: return False
        return True

    _constraints = [
        (_assignation_in_domain, 'Assignation out of domain!', ['datetime_from','datetime_to']),
        ]

    def _join(self, cr, uid, ids):
        query="""
        SELECT gac.id,gac.group_id,gac.datetime_to
        FROM groups_group_assignation AS ga
        INNER JOIN groups_group AS g ON ga.group_id=g.id
        LEFT JOIN groups_group AS gc ON g.parent_id=gc.parent_id AND g.classification=gc.classification
        LEFT JOIN groups_group_assignation AS gac ON gc.id=gac.group_id
            AND gac.participation_id=ga.participation_id
            AND (ga.datetime_to=gac.datetime_from OR ga.datetime_from=gac.datetime_to)
        WHERE ga.id in %(ids)s AND gac.id IS NOT NULL
        ORDER BY gac.datetime_from
        """
        cr.execute(query,{'ids': tuple(ids) })
        id_to_extends=0
        ids_to_unlink=[]
        groupy=None
        for (id, group_id, datetime_to) in cr.fetchall():
            if groupy and group_id==groupy:
                super(groups_group_assignation, self).write(cr, uid, [id_to_extends],
                    {'datetime_to': datetime_to})
                ids_to_unlink.append(id)
                break
            id_to_extends=id
            groupy=group_id
        super(groups_group_assignation, self).unlink(cr, uid, ids_to_unlink)

    def unlink(self, cr, uid, ids, context=None):
        if context is None:
            context = {}
        if not ids: return True
        context_unlinking=dict(context)
        context_unlinking.update({'unlinking':True})
        self.write(cr, uid, ids, {}, context=context_unlinking)
        self._join(cr, uid, ids)
        return super(groups_group_assignation, self).unlink(cr, uid, ids, context=context)

    def _divide(self, cr, uid, id_to_cut, datetime_cut, datetime_to, group_id, participation_id):
        children_ids=self.pool.get('groups.group').search(cr, uid, [('parent_id','=',group_id)])
        sub_ids_to_cut=self.search(cr, uid, [
            ('participation_id','=',participation_id),
            ('group_id','in',children_ids),
            ('datetime_from','<',datetime_cut),
            ('datetime_to','>',datetime_to)])
        for item in self.read(cr, uid, sub_ids_to_cut,
                ['group_id','participation_id','datetime_from','datetime_to'], context={'withoutBlank': False}):
            self._divide(cr, uid, item[id], datetime_cut, item['datetime_to'],
                group_id, participation_id)
        super(groups_group_assignation, self).write(cr, uid, id_to_cut, {'datetime_to': datetime_cut})
        super(groups_group_assignation, self).create(cr, uid, {
            'group_id': group_id,
            'participation_id': participation_id,
            'datetime_from': datetime_cut,
            'datetime_to': datetime_to})


    def create(self, cr, uid, vals, context=None):
        if 'datetime_from' not in vals: vals['datetime_from']=_MIN_DATETIME
        if 'datetime_to' not in vals: vals['datetime_to']=_MAX_DATETIME
        if not vals['datetime_from']: vals['datetime_from']=_MIN_DATETIME
        if not vals['datetime_to']: vals['datetime_to']=_MAX_DATETIME
        if context is None:
            context = {}
        """
        Si el contexte es extern cal mirar a on es crea la nova assignacio
        Si es crea enmig d'una altra assignacio de la mateixa classificacio,
            es retalla i a partir d'alla es crea una altra amb quedant la nova enmig.
            En aquest cas, si hi ha assignacions de subgrups, tambe s'han
            de retallar sense cap condicio.
        Si es crea al finalitzar o al comencar una altra assignacio de la mateixa classificacio, no es fa res.
        Despres, amb el write ja es fan la resta d'operacions.
        """
        group=self.pool.get('groups.group').browse(cr, uid, vals['group_id'], context=context)
        group_ids=[vals['group_id']]# No hem d'afegir interval buit si ja existeix
        if group.parent_id and group.classification:
            group_ids+=self.pool.get('groups.group').search(cr, uid, [
                ('parent_id','=',group.parent_id.id),
                ('classification','=',group.classification)])
        ids=self.search(cr, uid, [
                ('group_id','in',group_ids),
                ('participation_id','=',vals['participation_id']),
                ('datetime_from','<=',vals['datetime_from']),
                ('datetime_to','>=',vals['datetime_from'])])
        """
        Cal interval buit ? Si n'hi ha interval del mateix grup on comenca el
            nou interval que volem crear, no.
        Cal retallar cap interval ? Si no creem interval buit no. L'assignacio a
            retallar inclou exclusivament la data d'inici del nou interval
        """
        interval_buit=True
        ga_id=None
        id=None
        g_id=None
        new_context=dict(context)
        new_context['withoutBlank']=False
        for ga in self.read(cr, uid, ids, ['group_id','datetime_from','datetime_to'],
            context=new_context):
            if agafa_id(ga['group_id'])==vals['group_id']:
                interval_buit=False
                if vals['datetime_to']>ga['datetime_to']:
                    ga_id=ga['id']
                else:
                    return ga['id']
                break
            if ga['datetime_from']<vals['datetime_from']\
                or ga['datetime_to']>vals['datetime_from']:
                ga_id=ga['id']
                g_id=ga['group_id'][0]
                retallar_datetime_to=ga['datetime_to']
        if interval_buit and ga_id:
            self._divide(cr, uid, ga_id, vals['datetime_from'],
                retallar_datetime_to, g_id, vals['participation_id'])
        if not interval_buit and ga_id:
            ga_id=id
        else:
            vals2=dict(vals)
            vals2.update({'datetime_to': vals['datetime_from']})
            id=super(groups_group_assignation, self).create(cr, uid, vals2, context=context)
            context['interval_creation']=id

        self.write(cr, uid, [id], {'datetime_to': vals['datetime_to']}, context=context)
        return id

    def _extend_beside(self, cr, uid, data, context=None):
        ## Updating assignations -with same classification group- beside than we are changing
        ## Extending de beside assignations -with same classification group-
        if context is None:
            context = {}
        context['no_update_classification']=True
        
        for (id_up,participation_id,parent_id,classification,from_new,to_new,side) in data:
            if from_new==to_new: continue
            ids_to_extend_from_the_rigth=[]
            ids_to_extend_from_the_left=[]
            # Finding assignation from the left of same classification than we are changing
            if side in ('l','c'):
                ids_to_extend_from_the_rigth=self.search(cr, uid, [
                    ('group_id.classification','=',classification),
                    ('group_id.parent_id','=',parent_id),
                    ('participation_id','=',participation_id),
                    ('datetime_to','=',from_new)])
            # Finding assignation from the rigth of same classification than we are changing
            if side in ('r','c'):
                ids_to_extend_from_the_left=self.search(cr, uid, [
                    ('group_id.classification','=',classification),
                    ('group_id.parent_id','=',parent_id),
                    ('participation_id','=',participation_id),
                    ('datetime_from','=',to_new)])
            # Modifying found assignations
            if ids_to_extend_from_the_rigth and ids_to_extend_from_the_left and side=='c':
                # No side because we are deleting all the assignation. We must extend both sides
                dt_from_new=str_to_datetime(from_new)
                dt_to_new=str_to_datetime(to_new)
                td=dt_to_new-dt_from_new
                dt_half=None
                if td.days>1:
                    dt_half=dt_from_new+datetime.timedelta(days=int(td.days/2))
                else:
                    dt_half=dt_from_new+datetime.timedelta(seconds=int((td.days*3600*24+td.seconds)/2))
                half=dt_half.isoformat()
                self.write(cr, uid, ids_to_extend_from_the_rigth,
                    {'datetime_to': half}, context=context)
                self.write(cr, uid, ids_to_extend_from_the_left,
                    {'datetime_from': half}, context=context)
            else:
                # There are side to extends from
                if ids_to_extend_from_the_rigth:
                    # We extend from the rigth
                    self.write(cr, uid, ids_to_extend_from_the_rigth,
                        {'datetime_from': from_new}, context=context)
                if ids_to_extend_from_the_rigth:
                    # We extend from the left
                    self.write(cr, uid, ids_to_extend_from_the_rigth,
                        {'datetime_to': to_new}, context=context)

    def _erase_into(self, cr, uid, data, context=None):
        ## Deleting assignations -with same classification group- into new interval
        ## and cut than others who overlaps with enlargement into new interval
        if context is None:
            context = {}
        context['no_update_classification']=True

        same_classification_to_unlink=set()
        same_classification_to_change_datetime_to={}
        same_classification_to_change_datetime_from={}
        for (id_up,participation_id,parent_id,classification,from_new,to_new,side) in data:
            if from_new==to_new: continue
            # Finding assignations into new assignation interval
            conditions=[('group_id.classification','=',classification),
                        ('group_id.parent_id','=',parent_id),
                        ('participation_id','=',participation_id),
                        ('datetime_from','>=',from_new),
                        ('datetime_to','<=',to_new),]
            ids=self.search(cr, uid, conditions)
            if 'interval_creation' in context:
                if context['interval_creation'] in ids:
                    ids.remove(context['interval_creation'])
            same_classification_to_unlink.update(ids)
            if side in ('l','c'):
                same_classification_left=self.search(cr, uid, [
                        ('group_id.classification','=',classification),
                        ('group_id.parent_id','=',parent_id),
                        ('participation_id','=',participation_id),
                        ('datetime_from','<',from_new),
                        ('datetime_to','>=',from_new)])
                if same_classification_left:
                    same_classification_to_change_datetime_to.setdefault(from_new, []).\
                        append(same_classification_left[0])
            if side in ('r','c'):
                same_classification_right=self.search(cr, uid, [
                        ('group_id.classification','=',classification),
                        ('group_id.parent_id','=',parent_id),
                        ('participation_id','=',participation_id),
                        ('datetime_from','<=',to_new),
                        ('datetime_to','>',to_new)])
                if same_classification_right:
                    same_classification_to_change_datetime_from.setdefault(to_new, []).\
                        append(same_classification_right[0])

        if len(same_classification_to_unlink):
            self.unlink(cr, uid, list(same_classification_to_unlink), context=context)
        for (datetime_to,ids_to_change) in same_classification_to_change_datetime_to.iteritems():
            self.write(cr, uid, ids_to_change, {'datetime_to': datetime_to}, context=context)
        for (datetime_from,ids_to_change) in same_classification_to_change_datetime_from.iteritems():
            self.write(cr, uid, ids_to_change, {'datetime_from': datetime_from}, context=context)

    def _validate(self, cr, uid, ids, context=None):
        context = context or {}
        if context.get('no_check_constrains',False):
            return
        super(groups_group_assignation, self)._validate(cr, uid, ids, context=context)

    def _modify_children(self, cr, uid, to_unlink, to_modify_from, to_modify_to, from_new, to_new, context=None):
        ## Modifying assignations of child groups
        if context is None:
            context = {}
        context.pop('no_update_classification',False)
        self.unlink(cr, uid, list(to_unlink), context=context,)
        to_modify_from_and_to=to_modify_from.intersection(to_modify_to)
        if to_modify_from_and_to:
            self.write(cr, uid, list(to_modify_from_and_to), {
                'datetime_from': from_new,
                'datetime_to': to_new,
            }, context=dict(context.items()+[('no_check_constrains',True)]))
        if to_modify_from:
            self.write(cr, uid, list(to_modify_from.difference(to_modify_from_and_to)), {
                'datetime_from': from_new,
            }, context=dict(context.items()+[('no_check_constrains',True)]))
        if to_modify_to:
            self.write(cr, uid, list(to_modify_to.difference(to_modify_from_and_to)), {
                'datetime_to': to_new,
            }, context=dict(context.items()+[('no_check_constrains',True)]))

    def read(self, cr, uid, ids, fields, context=None, load='_classic_write'):
        if not fields: fields=[]
        if not context: context={}
        ret=super(groups_group_assignation, self).read(cr, uid, ids, fields, context=context, load=load)
        # @type context dict
        if context.get('withoutBlank', True):
            # @type ret list
            for item in ret:
                if 'datetime_from' in item:
                    if item['datetime_from']==_MIN_DATETIME: item['datetime_from']=''
                if 'datetime_to' in item:
                    if item['datetime_to']==_MAX_DATETIME: item['datetime_to']=''
        return ret

    def write(self, cr, uid, ids, vals, context=None):
        if not vals: vals={}
        if not ids: return True
        if context is None:
            context = {}
            
        # Default value for variables
        if 'datetime_from' in vals:
            if not vals['datetime_from'] or vals['datetime_from']<_MIN_DATETIME:
                vals['datetime_from']=_MIN_DATETIME
        if 'datetime_to' in vals:
            if not vals['datetime_to'] or vals['datetime_to']>_MAX_DATETIME:
                vals['datetime_to']=_MAX_DATETIME

        # Changing interval group assignation
        if vals.has_key('datetime_from') or vals.has_key('datetime_to') or context.get('unlinking',False):

            to_unlink=set()
            to_modify_from=set()
            to_modify_to=set()
            to_delete=set()
            to_create=set()
            from_new=None
            to_new=None
            query="""
            SELECT
                gad.id AS id_down,gad.datetime_from AS dt_from_down,gad.datetime_to AS dt_to_down,
                ga.id AS id_up,ga.datetime_from AS dt_from_up,ga.datetime_to AS dt_to_up,
                ga.participation_id,g.classification,g.parent_id
            FROM groups_group_assignation AS ga
                LEFT JOIN groups_group AS g ON ga.group_id=g.id
                LEFT JOIN groups_group AS gd ON gd.parent_id=g.id
                LEFT JOIN groups_group_assignation AS gad
                   ON gd.id=gad.group_id AND ga.participation_id=gad.participation_id
            WHERE ga.id IN %s
            """
            cr.execute(query,(tuple(ids),))
            for (id_down,from_down,to_down,id_up,from_up,to_up,
                    participation_id,classification,parent_id) in cr.fetchall():
                (from_new,to_new)=(from_up,to_up)
                if 'datetime_from' in vals:
                    from_new=vals['datetime_from']
                if 'datetime_to' in vals:
                    to_new=vals['datetime_to']
                if context.get('unlinking',False): to_new=from_up
                ## Preparing data to modify assignations of group children
                if id_down:
                    if id_down in ids:
                        if from_new!=from_up: from_down=from_new
                        if to_new!=to_up: to_down=to_new
                    if from_new!=from_up:
                        if to_down<=from_new:
                            to_unlink.add(id_down)
                        else:
                            if from_new>from_down or (from_up==from_down and from_new<from_up):
                                to_modify_from.add(id_down)
                    if to_new!=to_up:
                        if from_down>=to_new:
                            to_unlink.add(id_down)
                        else:
                            if to_new<to_down or (to_up==to_down and to_new>to_up):
                                to_modify_to.add(id_down)
                    if from_new!=from_up and to_new!=to_up:
                        if (from_new>to_up and to_down==to_up) or \
                            (to_new<from_up and from_down==from_up):
                            to_modify_to.add(id_down)
                            to_modify_from.add(id_down)
                ## Preparing data to modify groups with same classification affected by the change
                if classification and parent_id and not context.get('no_update_classification',False):
                    key=(id_up,participation_id,parent_id,classification)
                    if from_new>to_up or to_new<from_up:
                        to_delete.add( key + (from_up,to_up,'c') )
                        to_create.add( key + (from_new,to_new,'c') )
                    if from_new>=from_up and to_new<to_up and from_up!=to_up:
                        to_delete.add( key + (from_up,from_new,'l') )
                        to_delete.add( key + (to_new,to_up,'r') )
                    if from_new>from_up:
                        to_delete.add( key + (from_up,from_new,'l') )
                    else:
                        to_create.add( key + (from_new,from_up,'l') )
                    if to_new<to_up:
                        to_delete.add( key + (to_new,to_up,'r') )
                    else:
                        to_create.add( key + (to_up,to_new,'r') )

            self._extend_beside(cr, uid, to_delete, context=context)
            self._erase_into(cr, uid, to_create, context=context)
            self._modify_children(cr, uid, to_unlink, to_modify_from,
                to_modify_to, from_new, to_new, context=context)

        ret=super(groups_group_assignation, self).write(cr, uid, ids, vals, context=context)
        return ret

groups_group_assignation()

class groups_participation(osv.osv):
    _name = 'groups.participation'

    _description = 'A participation in groups.'


    _columns = {
        'name' : fields.char('Ref',size=32),
        'participant' : fields.many2one('res.partner.contact', 'Participant',
            required=True, ondelete="cascade", select=1,
            help="The contact of the participation"),
        'assignation_ids' : fields.one2many('groups.group_assignation',
            'participation_id','Assignations',),
    }

    _defaults = {
        'name': lambda obj, cr, uid, context:
            obj.pool.get('ir.sequence').get(cr, uid, 'groups.ref.participation'),
    }

    def groups_in_interval(self, cr, uid, ids, date_from=_MIN_DATETIME, date_to=_MAX_DATETIME, context=None):
        """
        Retorna un diccionari amb claus els identificadors de les participacions
        contingudes en la llista del paràmetre ids. Els valors són una llista de
        diccionaris amb dades sobre els grups amb assignacions per la participació
        identificada amb la clau dins el periode determinat pels paràmetres
        date_from i date_to. La llista de grups es troba ordenada per prioritat
        éssent el grup amb prioritat més alta el quin és primer.
        """
        if not ids: return {}
        res={}
        for id in ids: res[id]=[]
        if ids:
            query="""
            SELECT ga.participation_id,ga.datetime_from,ga.datetime_to,g.id AS gid,g.priority
            FROM groups_group_assignation AS ga
            INNER JOIN groups_group AS g ON ga.group_id=g.id
            WHERE ga.participation_id in %(part_ids)s
                AND ga.limit_from<'%(date_to)s'
                AND ga.limit_to>'%(date_from)s'
            ORDER BY g.priority,ga.datetime_from DESC
            """ % {'part_ids': tuple(ids+[0]),'date_from': date_from, 'date_to': date_to, }
            cr.execute(query)
            for (part_id,datetime_from,datetime_to,group_id,priority) in cr.fetchall():
                if part_id not in res: res[part_id]=[]
                res[part_id]+=[{'group_id': group_id,'datetime_from': datetime_from,'datetime_to': datetime_to,'priority':priority}]
        return res

groups_participation()

class groups_group(osv.osv):
    _name = 'groups.group'
    _description = 'A group.'
    _parent_store = True

    _columns = {
        'name' : fields.char('Name', size=32, required=True, select=1, help="The group name"),
        'parent_id' : fields.many2one('groups.group','Parent',required=False,
            select=1, ondelete='cascade',help="The parent group"),
        'parent_left' : fields.integer('parent_left',),
        'parent_right' : fields.integer('parent_right',),
        'assignation_ids' : fields.one2many('groups.group_assignation','group_id','Assignations'),
        'children_ids' : fields.one2many('groups.group','parent_id',help="The subgroups of the group"),
        'classification' : fields.char('Classification',size=30,),
        'priority' : fields.integer('Priority'),
    }

    def _check_recursion(self, cr, uid, ids):
        level = 100
        while len(ids):
            cr.execute('select distinct parent_id from groups_group where id in ('+
                ','.join(map(str,ids))+')')
            ids = filter(None, map(lambda x:x[0], cr.fetchall()))
            if not level:
                return False
            level -= 1
        return True

    _constraints = [
        (_check_recursion, 'Error ! You can not create recursive groups.', ['parent_id'])
    ]


    def _domain_assignations(self, cr, uid, ids, context=None):
        ret={}
        domains={}
        for item in self.browse(cr, uid, ids, context={'withoutBlank': False}):
            id=item.id
            if item.parent_id:
                pid=item.parent_id.id
                if not pid in domains:
                    dict={}
                    for assignation in item.parent_id.assignation_ids:
                        key=assignation.participation_id.id
                        value=(assignation.datetime_from,assignation.datetime_to)
                        if not key in dict:
                            dict[key]=[value]
                        else:
                            dict[key].append(value)
                    domains[pid]=dict
                ret[id]=domains[pid]
            else:
                ret[id]=None # No parent, no domain
        return ret

groups_group()
